# lutris_source.py
#
# Copyright 2022-2024 kramo
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from shutil import rmtree
from sqlite3 import connect
from typing import NamedTuple

from cartridges import shared
from cartridges.game import Game
from cartridges.importer.location import Location, LocationSubPath
from cartridges.importer.source import SourceIterable, URLExecutableSource
from cartridges.utils.sqlite import copy_db


class LutrisSourceIterable(SourceIterable):
    source: "LutrisSource"

    def __iter__(self):
        """Generator method producing games"""

        # Query the database
        request = """
            SELECT
                games.id,
                games.name,
                games.slug,
                games.runner,
                categories.name = ".hidden" as hidden
            FROM
                games
            LEFT JOIN
                games_categories ON games_categories.game_id = games.id
            FULL JOIN
                categories ON games_categories.category_id = categories.id
            WHERE
                games.name IS NOT NULL
                AND games.slug IS NOT NULL
                AND games.configPath IS NOT NULL
                AND games.installed
                AND (games.runner IS NOT "steam" OR :import_steam)
                AND (games.runner IS NOT "flatpak" OR :import_flatpak)
            ;
        """

        params = {
            "import_steam": shared.schema.get_boolean("lutris-import-steam"),
            "import_flatpak": shared.schema.get_boolean("lutris-import-flatpak"),
        }
        db_path = copy_db(self.source.locations.data["pga.db"])
        connection = connect(db_path)
        cursor = connection.execute(request, params)
        coverart_is_dir = (
            coverart_path := self.source.locations.data.root / "coverart"
        ).is_dir()

        # Create games from the DB results
        for row in cursor:
            # Create game
            values = {
                "added": shared.import_time,
                "hidden": row[4],
                "name": row[1],
                "source": f"{self.source.source_id}_{row[3]}",
                "game_id": self.source.game_id_format.format(
                    runner=row[3], game_id=row[0]
                ),
                "executable": self.source.make_executable(game_id=row[0]),
            }
            game = Game(values)
            additional_data = {}

            # Get official image path
            if coverart_is_dir:
                image_path = coverart_path / f"{row[2]}.jpg"
                additional_data["local_image_path"] = image_path

            yield (game, additional_data)

        # Cleanup
        rmtree(str(db_path.parent))


class LutrisLocations(NamedTuple):
    data: Location


class LutrisSource(URLExecutableSource):
    """Generic Lutris source"""

    source_id = "lutris"
    name = _("Lutris")
    iterable_class = LutrisSourceIterable
    url_format = "lutris:rungameid/{game_id}"
    available_on = {"linux"}

    locations: LutrisLocations

    @property
    def game_id_format(self):
        return self.source_id + "_{runner}_{game_id}"

    def __init__(self) -> None:
        super().__init__()
        self.locations = LutrisLocations(
            Location(
                schema_key="lutris-location",
                candidates=(
                    shared.flatpak_dir / "net.lutris.Lutris" / "data" / "lutris",
                    shared.data_dir / "lutris",
                    shared.host_data_dir / "lutris",
                ),
                paths={
                    "pga.db": LocationSubPath("pga.db"),
                },
                invalid_subtitle=Location.DATA_INVALID_SUBTITLE,
            )
        )



from typing import Iterable, Optional


class FriendlyError(Exception):
    """
    An error that is supposed to be shown to the user in a nice format

    Use `raise ... from ...` to preserve context.
    """

    title_format: str
    title_args: Iterable[str]
    subtitle_format: str
    subtitle_args: Iterable[str]

    @property
    def title(self) -> str:
        """Get the gettext translated error title"""
        return self.title_format.format(self.title_args)

    @property
    def subtitle(self) -> str:
        """Get the gettext translated error subtitle"""
        return self.subtitle_format.format(self.subtitle_args)

    def __init__(
        self,
        title: str,
        subtitle: str,
        title_args: Optional[Iterable[str]] = None,
        subtitle_args: Optional[Iterable[str]] = None,
    ) -> None:
        """Create a friendly error

        :param str title: The error's title, translatable with gettext
        :param str subtitle: The error's subtitle, translatable with gettext
        """
        super().__init__()
        if title is not None:
            self.title_format = title
        if subtitle is not None:
            self.subtitle_format = subtitle
        self.title_args = title_args if title_args else ()
        self.subtitle_args = subtitle_args if subtitle_args else ()

    def __str__(self) -> str:
        return f"{self.title} - {self.subtitle}"



# pipeline.py
#
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from typing import Iterable

from gi.repository import GObject

from cartridges.game import Game
from cartridges.store.managers.manager import Manager


class Pipeline(GObject.Object):
    """Class representing a set of managers for a game"""

    game: Game
    additional_data: dict

    waiting: set[Manager]
    running: set[Manager]
    done: set[Manager]

    def __init__(
        self, game: Game, additional_data: dict, managers: Iterable[Manager]
    ) -> None:
        super().__init__()
        self.game = game
        self.additional_data = additional_data
        self.waiting = set(managers)
        self.running = set()
        self.done = set()

    @property
    def not_done(self) -> set[Manager]:
        """Get the managers that are not done yet"""
        return self.waiting | self.running

    @property
    def is_done(self) -> bool:
        return len(self.waiting) == 0 and len(self.running) == 0

    @property
    def blocked(self) -> set[Manager]:
        """Get the managers that cannot run because their dependencies aren't done"""
        blocked = set()
        for waiting in self.waiting:
            for not_done in self.not_done:
                if waiting == not_done:
                    continue
                if type(not_done) in waiting.run_after:
                    blocked.add(waiting)
        return blocked

    @property
    def ready(self) -> set[Manager]:
        """Get the managers that can be run"""
        return self.waiting - self.blocked

    @property
    def progress(self) -> float:
        """Get the pipeline progress. Should only be a rough idea."""
        n_done = len(self.done)
        n_total = len(self.waiting) + len(self.running) + n_done
        try:
            progress = n_done / n_total
        except ZeroDivisionError:
            progress = 1
        return progress

    def advance(self) -> None:
        """Spawn tasks for managers that are able to run for a game"""

        # Separate blocking / async managers
        managers = self.ready
        blocking = set(filter(lambda manager: manager.blocking, managers))
        parallel = managers - blocking

        # Schedule parallel managers, then run the blocking ones
        for manager in (*parallel, *blocking):
            self.waiting.remove(manager)
            self.running.add(manager)
            manager.process_game(self.game, self.additional_data, self.manager_callback)

    def manager_callback(self, manager: Manager) -> None:
        """Method called by a manager when it's done"""
        logging.debug("%s done for %s", manager.name, self.game.game_id)
        self.running.remove(manager)
        self.done.add(manager)
        self.emit("advanced")
        self.advance()

    @GObject.Signal(name="advanced")
    def advanced(self):  # type: ignore
        """Signal emitted when the pipeline has advanced"""



# sqlite.py
#
# Copyright 2022-2023 kramo
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from glob import escape
from pathlib import Path
from shutil import copyfile

from gi.repository import GLib


def copy_db(original_path: Path) -> Path:
    """
    Copy a sqlite database to a cache dir and return its new path.
    The caller in in charge of deleting the returned path's parent dir.
    """
    tmp = Path(GLib.Dir.make_tmp())
    for file in original_path.parent.glob(f"{escape(original_path.name)}*"):
        copy = tmp / file.name
        copyfile(str(file), str(copy))
    return tmp / original_path.name



# sgdb_manager.py
#
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from json import JSONDecodeError

from requests.exceptions import HTTPError, SSLError

from cartridges.errors.friendly_error import FriendlyError
from cartridges.game import Game
from cartridges.store.managers.async_manager import AsyncManager
from cartridges.store.managers.cover_manager import CoverManager
from cartridges.store.managers.steam_api_manager import SteamAPIManager
from cartridges.utils.steamgriddb import SgdbAuthError, SgdbHelper


class SgdbManager(AsyncManager):
    """Manager in charge of downloading a game's cover from SteamGridDB"""

    run_after = (SteamAPIManager, CoverManager)
    retryable_on = (HTTPError, SSLError, ConnectionError, JSONDecodeError)

    def main(self, game: Game, _additional_data: dict) -> None:
        try:
            sgdb = SgdbHelper()
            sgdb.conditionaly_update_cover(game)
        except SgdbAuthError as error:
            # If invalid auth, cancel all SGDBManager tasks
            self.cancellable.cancel()
            raise FriendlyError(
                _("Couldn't Authenticate SteamGridDB"),
                _("Verify your API key in preferences"),
            ) from error



# color_log_formatter.py
#
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import Formatter, LogRecord


class ColorLogFormatter(Formatter):
    """Formatter that outputs logs in a colored format"""

    RESET = "\033[0m"
    DIM = "\033[2m"
    BOLD = "\033[1m"
    RED = "\033[31m"
    YELLOW = "\033[33m"

    def format(self, record: LogRecord) -> str:
        super_format = super().format(record)
        match record.levelname:
            case "CRITICAL":
                return self.BOLD + self.RED + super_format + self.RESET
            case "ERROR":
                return self.RED + super_format + self.RESET
            case "WARNING":
                return self.YELLOW + super_format + self.RESET
            case "DEBUG":
                return self.DIM + super_format + self.RESET
            case _other:
                return super_format



# heroic_source.py
#
# Copyright 2022-2023 kramo
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import logging
from abc import abstractmethod
from functools import cached_property
from hashlib import sha256
from json import JSONDecodeError
from pathlib import Path
from typing import Iterable, NamedTuple, Optional, TypedDict

from cartridges import shared
from cartridges.game import Game
from cartridges.importer.location import Location, LocationSubPath
from cartridges.importer.source import (
    SourceIterable,
    SourceIterationResult,
    URLExecutableSource,
)


def path_json_load(path: Path):
    """
    Load JSON from the file at the given path

    :raises OSError: if the file can't be opened
    :raises JSONDecodeError: if the file isn't valid JSON
    """
    with path.open("r", encoding="utf-8") as open_file:
        return json.load(open_file)


class InvalidLibraryFileError(Exception):
    pass


class InvalidInstalledFileError(Exception):
    pass


class InvalidStoreFileError(Exception):
    pass


class HeroicLibraryEntry(TypedDict):
    app_name: str
    installed: Optional[bool]
    runner: str
    title: str
    developer: str
    art_square: str


class SubSourceIterable(Iterable):
    """Class representing a Heroic sub-source"""

    source: "HeroicSource"
    source_iterable: "HeroicSourceIterable"
    name: str
    service: str
    image_uri_params: str = ""
    relative_library_path: Path
    library_json_entries_key: str = "library"

    def __init__(self, source, source_iterable) -> None:
        self.source = source
        self.source_iterable = source_iterable

    @cached_property
    def library_path(self) -> Path:
        path = self.source.locations.config.root / self.relative_library_path
        logging.debug("Using Heroic %s library.json path %s", self.name, path)
        return path

    def process_library_entry(self, entry: HeroicLibraryEntry) -> SourceIterationResult:
        """Build a Game from a Heroic library entry"""

        app_name = entry["app_name"]
        runner = entry["runner"]

        # Build game
        values = {
            "source": f"{self.source.source_id}_{self.service}",
            "added": shared.import_time,
            "name": entry["title"],
            "developer": entry.get("developer", None),
            "game_id": self.source.game_id_format.format(
                service=self.service, game_id=app_name
            ),
            "executable": self.source.make_executable(runner=runner, app_name=app_name),
            "hidden": self.source_iterable.is_hidden(app_name),
        }
        game = Game(values)

        # Get the image path from the Heroic cache
        # Filenames are derived from the URL that Heroic used to get the file
        uri: str = entry["art_square"] + self.image_uri_params
        digest = sha256(uri.encode()).hexdigest()
        image_path = self.source.locations.config.root / "images-cache" / digest
        additional_data = {"local_image_path": image_path}

        return (game, additional_data)

    def __iter__(self):
        """
        Iterate through the games with a generator
        :raises InvalidLibraryFileError: on initial call if the library file is bad
        """

        try:
            iterator = iter(
                path_json_load(self.library_path)[self.library_json_entries_key]
            )
        except (OSError, JSONDecodeError, TypeError, KeyError) as error:
            raise InvalidLibraryFileError(
                f"Invalid {self.library_path.name}"
            ) from error
        for entry in iterator:
            try:
                yield self.process_library_entry(entry)
            except KeyError as error:
                logging.warning(
                    "Skipped invalid %s game %s",
                    self.name,
                    entry.get("app_name", "UNKNOWN"),
                    exc_info=error,
                )
                continue


class StoreSubSourceIterable(SubSourceIterable):
    """
    Class representing a "store" sub source.
    Games can be installed or not, this class does the check accordingly.
    """

    relative_installed_path: Path
    installed_app_names: set[str]

    @cached_property
    def installed_path(self) -> Path:
        path = self.source.locations.config.root / self.relative_installed_path
        logging.debug("Using Heroic %s installed.json path %s", self.name, path)
        return path

    @abstractmethod
    def get_installed_app_names(self) -> set[str]:
        """
        Get the sub source's installed app names as a set.

        :raises InvalidInstalledFileError: if the installed file data cannot be read
            Whenever possible, `__cause__` is set with the original exception
        """

    def is_installed(self, app_name: str) -> bool:
        return app_name in self.installed_app_names

    def process_library_entry(self, entry):
        # Skip games that are not installed
        app_name = entry["app_name"]
        if not self.is_installed(app_name):
            logging.warning(
                "Skipped %s game %s (%s): not installed",
                self.service,
                entry["title"],
                app_name,
            )
            return None
        # Process entry as normal
        return super().process_library_entry(entry)

    def __iter__(self):
        """
        Iterate through the installed games with a generator
        :raises InvalidLibraryFileError: on initial call if the library file is bad
        :raises InvalidInstalledFileError: on initial call if the installed file is bad
        """
        self.installed_app_names = self.get_installed_app_names()
        yield from super().__iter__()


class SideloadIterable(SubSourceIterable):
    name = "sideload"
    service = "sideload"
    relative_library_path = Path("sideload_apps") / "library.json"
    library_json_entries_key = "games"


class LegendaryIterable(StoreSubSourceIterable):
    name = "legendary"
    service = "epic"
    image_uri_params = "?h=400&resize=1&w=300"
    relative_library_path = Path("store_cache") / "legendary_library.json"

    # relative_installed_path = (
    #    Path("legendary") / "legendaryConfig" / "legendary" / "installed.json"
    # )

    @cached_property
    def installed_path(self) -> Path:
        """Get the right path depending on the Heroic version"""
        # TODO after Heroic 2.9 has been out for a while
        # We should use the commented out relative_installed_path
        # and remove this property override.

        heroic_config_path = self.source.locations.config.root
        # Heroic >= 2.9
        if (path := heroic_config_path / "legendaryConfig").is_dir():
            logging.debug("Using Heroic >= 2.9 legendary file")
        # Heroic <= 2.8
        elif heroic_config_path.is_relative_to(shared.flatpak_dir):
            # Heroic flatpak
            path = shared.flatpak_dir / "com.heroicgameslauncher.hgl" / "config"
            logging.debug("Using Heroic flatpak <= 2.8 legendary file")
        else:
            # Heroic native
            logging.debug("Using Heroic native <= 2.8 legendary file")
            path = shared.host_config_dir

        path = path / "legendary" / "installed.json"
        logging.debug("Using Heroic %s installed.json path %s", self.name, path)
        return path

    def get_installed_app_names(self):
        try:
            return set(path_json_load(self.installed_path).keys())
        except (OSError, JSONDecodeError, AttributeError) as error:
            raise InvalidInstalledFileError(
                f"Invalid {self.installed_path.name}"
            ) from error


class GogIterable(StoreSubSourceIterable):
    name = "gog"
    service = "gog"
    library_json_entries_key = "games"
    relative_library_path = Path("store_cache") / "gog_library.json"
    relative_installed_path = Path("gog_store") / "installed.json"

    def get_installed_app_names(self):
        try:
            return {
                app_name
                for entry in path_json_load(self.installed_path)["installed"]
                if (app_name := entry.get("appName")) is not None
            }
        except (OSError, JSONDecodeError, KeyError, AttributeError) as error:
            raise InvalidInstalledFileError(
                f"Invalid {self.installed_path.name}"
            ) from error


class NileIterable(StoreSubSourceIterable):
    name = "nile"
    service = "amazon"
    relative_library_path = Path("store_cache") / "nile_library.json"
    relative_installed_path = Path("nile_config") / "nile" / "installed.json"

    def get_installed_app_names(self):
        try:
            installed_json = path_json_load(self.installed_path)
            return {
                app_name
                for entry in installed_json
                if (app_name := entry.get("id")) is not None
            }
        except (OSError, JSONDecodeError, AttributeError) as error:
            raise InvalidInstalledFileError(
                f"Invalid {self.installed_path.name}"
            ) from error


class HeroicSourceIterable(SourceIterable):
    source: "HeroicSource"

    hidden_app_names: set[str] = set()

    def is_hidden(self, app_name: str) -> bool:
        return app_name in self.hidden_app_names

    def get_hidden_app_names(self) -> set[str]:
        """Get the hidden app names from store/config.json

        :raises InvalidStoreFileError: if the store is invalid for some reason
        """

        try:
            store = path_json_load(self.source.locations.config["store_config.json"])
            self.hidden_app_names = {
                app_name
                for game in store["games"]["hidden"]
                if (app_name := game.get("appName")) is not None
            }
        except KeyError:
            logging.warning('No ["games"]["hidden"] key in Heroic store file')
        except (OSError, JSONDecodeError, TypeError) as error:
            logging.error("Invalid Heroic store file", exc_info=error)
            raise InvalidStoreFileError() from error

    def __iter__(self):
        """Generator method producing games from all the Heroic sub-sources"""

        self.get_hidden_app_names()

        # Get games from the sub sources
        for sub_source_class in (
            SideloadIterable,
            LegendaryIterable,
            GogIterable,
            NileIterable,
        ):
            sub_source = sub_source_class(self.source, self)

            if not shared.schema.get_boolean("heroic-import-" + sub_source.service):
                logging.debug("Skipping Heroic %s: disabled", sub_source.service)
                continue
            try:
                sub_source_iterable = iter(sub_source)
                yield from sub_source_iterable
            except (InvalidLibraryFileError, InvalidInstalledFileError) as error:
                logging.error(
                    "Skipping bad Heroic sub-source %s",
                    sub_source.service,
                    exc_info=error,
                )
                continue


class HeroicLocations(NamedTuple):
    config: Location


class HeroicSource(URLExecutableSource):
    """Generic Heroic Games Launcher source"""

    source_id = "heroic"
    name = _("Heroic")
    iterable_class = HeroicSourceIterable
    url_format = "heroic://launch/{runner}/{app_name}"
    available_on = {"linux", "win32", "darwin"}

    locations: HeroicLocations

    @property
    def game_id_format(self) -> str:
        """The string format used to construct game IDs"""
        return self.source_id + "_{service}_{game_id}"

    def __init__(self) -> None:
        super().__init__()
        self.locations = HeroicLocations(
            Location(
                schema_key="heroic-location",
                candidates=(
                    shared.config_dir / "heroic",
                    shared.host_config_dir / "heroic",
                    shared.flatpak_dir
                    / "com.heroicgameslauncher.hgl"
                    / "config"
                    / "heroic",
                    shared.appdata_dir / "heroic",
                    shared.app_support_dir / "heroic",
                ),
                paths={
                    "config.json": LocationSubPath("config.json"),
                    "store_config.json": LocationSubPath("store/config.json"),
                },
                invalid_subtitle=Location.CONFIG_INVALID_SUBTITLE,
            )
        )



# manager.py
#
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from abc import abstractmethod
from time import sleep
from typing import Any, Callable, Container

from cartridges.errors.error_producer import ErrorProducer
from cartridges.errors.friendly_error import FriendlyError
from cartridges.game import Game


class Manager(ErrorProducer):
    """Class in charge of handling a post creation action for games.

    * May connect to signals on the game to handle them.
    * May cancel its running tasks on critical error,
    in that case a new cancellable must be generated for new tasks to run.
    * May be retried on some specific error types
    """

    run_after: Container[type["Manager"]] = tuple()
    blocking: bool = True

    retryable_on: Container[type[Exception]] = tuple()
    continue_on: Container[type[Exception]] = tuple()
    signals: Container[type[str]] = set()
    retry_delay: int = 3
    max_tries: int = 3

    @property
    def name(self) -> str:
        return type(self).__name__

    @abstractmethod
    def main(self, game: Game, additional_data: dict) -> None:
        """
        Manager specific logic triggered by the run method
        * Implemented by final child classes
        * May block its thread
        * May raise retryable exceptions that will trigger a retry if possible
        * May raise other exceptions that will be reported
        """

    def run(self, game: Game, additional_data: dict) -> None:
        """Handle errors (retry, ignore or raise) that occur in the manager logic"""

        # Keep track of the number of tries
        tries = 1

        def handle_error(error: Exception) -> None:
            nonlocal tries

            # If FriendlyError, handle its cause instead
            base_error = error
            if isinstance(error, FriendlyError):
                error = error.__cause__

            log_args = (
                type(error).__name__,
                self.name,
                f"{game.name} ({game.game_id})",
            )

            out_of_retries_format = "Out of retries dues to %s in %s for %s"
            retrying_format = "Retrying %s in %s for %s"
            unretryable_format = "Unretryable %s in %s for %s"

            if type(error) in self.continue_on:
                # Handle skippable errors (skip silently)
                return

            if type(error) in self.retryable_on:
                if tries > self.max_tries:
                    # Handle being out of retries
                    logging.error(out_of_retries_format, *log_args)
                    self.report_error(base_error)
                else:
                    # Handle retryable errors
                    logging.error(retrying_format, *log_args)
                    sleep(self.retry_delay)
                    tries += 1
                    try_manager_logic()

            else:
                # Handle unretryable errors
                logging.error(unretryable_format, *log_args, exc_info=error)
                self.report_error(base_error)

        def try_manager_logic() -> None:
            try:
                self.main(game, additional_data)
            except Exception as error:  # pylint: disable=broad-exception-caught
                handle_error(error)

        try_manager_logic()

    def process_game(
        self, game: Game, additional_data: dict, callback: Callable[["Manager"], Any]
    ) -> None:
        """Pass the game through the manager"""
        self.run(game, additional_data)
        callback(self)



# local_cover_manager.py
#
# Copyright 2023 Geoffrey Coulaud
# Copyright 2023 kramo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from pathlib import Path
from typing import NamedTuple, Optional

import requests
from gi.repository import GdkPixbuf, Gio, GLib
from requests.exceptions import HTTPError, SSLError

from cartridges import shared
from cartridges.game import Game
from cartridges.store.managers.manager import Manager
from cartridges.store.managers.steam_api_manager import SteamAPIManager
from cartridges.utils.save_cover import convert_cover, save_cover


class ImageSize(NamedTuple):
    width: float = 0
    height: float = 0

    @property
    def aspect_ratio(self) -> float:
        return self.width / self.height

    def __str__(self):
        return f"{self.width}x{self.height}"

    def __mul__(self, scale: float | int) -> "ImageSize":
        return ImageSize(
            self.width * scale,
            self.height * scale,
        )

    def __truediv__(self, divisor: float | int) -> "ImageSize":
        return self * (1 / divisor)

    def __add__(self, other_size: "ImageSize") -> "ImageSize":
        return ImageSize(
            self.width + other_size.width,
            self.height + other_size.height,
        )

    def __sub__(self, other_size: "ImageSize") -> "ImageSize":
        return self + (other_size * -1)

    def element_wise_div(self, other_size: "ImageSize") -> "ImageSize":
        """Divide every element of self by the equivalent in the other size"""
        return ImageSize(
            self.width / other_size.width,
            self.height / other_size.height,
        )

    def element_wise_mul(self, other_size: "ImageSize") -> "ImageSize":
        """Multiply every element of self by the equivalent in the other size"""
        return ImageSize(
            self.width * other_size.width,
            self.height * other_size.height,
        )

    def invert(self) -> "ImageSize":
        """Invert the element of self"""
        return ImageSize(1, 1).element_wise_div(self)


class CoverManager(Manager):
    """
    Manager in charge of adding the cover image of the game

    Order of priority is:
    1. local cover
    2. icon cover
    3. online cover
    """

    run_after = (SteamAPIManager,)
    retryable_on = (HTTPError, SSLError, ConnectionError)

    def download_image(self, url: str) -> Path:
        image_file = Gio.File.new_tmp()[0]
        path = Path(image_file.get_path())
        with requests.get(url, timeout=5) as cover:
            cover.raise_for_status()
            path.write_bytes(cover.content)
        return path

    def is_stretchable(self, source_size: ImageSize, cover_size: ImageSize) -> bool:
        is_taller = source_size.aspect_ratio < cover_size.aspect_ratio
        if is_taller:
            return True
        max_stretch = 0.12
        resized_height = (1 / source_size.aspect_ratio) * cover_size.width
        stretch = 1 - (resized_height / cover_size.height)
        return stretch <= max_stretch

    def composite_cover(
        self,
        image_path: Path,
        scale: float = 1,
        blur_size: ImageSize = ImageSize(2, 2),
    ) -> GdkPixbuf.Pixbuf:
        """
        Return the image composited with a background blur.
        If the image is stretchable, just stretch it.

        :param path: Path where the source image is located
        :param scale:
            Scale of the smalled image side
            compared to the corresponding side in the cover
        :param blur_size: Size of the downscaled image used for the blur
        """

        # Load source image
        try:
            source = GdkPixbuf.Pixbuf.new_from_file(
                str(image_path)
            )
        except GLib.Error:
            tmp_cover_path = convert_cover(image_path, resize=False)

            if tmp_cover_path:
                source = GdkPixbuf.Pixbuf.new_from_file(
                    str(tmp_cover_path)
                )
                tmp_cover_path.unlink(missing_ok=True)
            else:
                return None

        source_size = ImageSize(source.get_width(), source.get_height())
        cover_size = ImageSize._make(shared.image_size)

        # Stretch if possible
        if scale == 1 and self.is_stretchable(source_size, cover_size):
            return source

        # Create the blurred cover background
        # fmt: off
        cover = (
            source
            .scale_simple(*blur_size, GdkPixbuf.InterpType.BILINEAR)
            .scale_simple(*cover_size, GdkPixbuf.InterpType.BILINEAR)
        )
        # fmt: on

        # Scale to fit, apply scaling, then center
        uniform_scale = scale * min(cover_size.element_wise_div(source_size))
        source_in_cover_size = source_size * uniform_scale
        source_in_cover_position = (cover_size - source_in_cover_size) / 2

        # Center the scaled source image in the cover
        source.composite(
            cover,
            *source_in_cover_position,
            *source_in_cover_size,
            *source_in_cover_position,
            uniform_scale,
            uniform_scale,
            GdkPixbuf.InterpType.BILINEAR,
            255,
        )
        return cover

    def main(self, game: Game, additional_data: dict) -> None:
        if game.blacklisted:
            return
        for key in (
            "local_image_path",
            "local_icon_path",
            "online_cover_url",
        ):
            # Get an image path
            if not (value := additional_data.get(key)):
                continue
            if key == "online_cover_url":
                image_path = self.download_image(value)
            else:
                image_path = Path(value)
            if not image_path.is_file():
                continue

            # Icon cover
            composite_kwargs = {}

            if key == "local_icon_path":
                composite_kwargs["scale"] = 0.7
                composite_kwargs["blur_size"] = ImageSize(1, 2)

            save_cover(
                game.game_id,
                pixbuf=self.composite_cover(image_path, **composite_kwargs),
            )

            if key == "online_cover_url":
                image_path.unlink(missing_ok=True)



# session_file_handler.py
#
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import lzma
from io import TextIOWrapper
from logging import StreamHandler
from lzma import FORMAT_XZ, PRESET_DEFAULT
from os import PathLike
from pathlib import Path
from typing import Optional

from cartridges import shared


class SessionFileHandler(StreamHandler):
    """
    A logging handler that writes to a new file on every app restart.
    The files are compressed and older sessions logs are kept up to a small limit.
    """

    NUMBER_SUFFIX_POSITION = 1

    backup_count: int
    filename: Path
    log_file: Optional[TextIOWrapper] = None

    def create_dir(self) -> None:
        """Create the log dir if needed"""
        self.filename.parent.mkdir(exist_ok=True, parents=True)

    def path_is_logfile(self, path: Path) -> bool:
        return path.is_file() and path.name.startswith(self.filename.stem)

    def path_has_number(self, path: Path) -> bool:
        try:
            int(path.suffixes[self.NUMBER_SUFFIX_POSITION][1:])
        except (ValueError, IndexError):
            return False
        return True

    def get_path_number(self, path: Path) -> int:
        """Get the number extension in the filename as an int"""
        suffixes = path.suffixes
        number = (
            0
            if not self.path_has_number(path)
            else int(suffixes[self.NUMBER_SUFFIX_POSITION][1:])
        )
        return number

    def set_path_number(self, path: Path, number: int) -> str:
        """Set or add the number extension in the filename"""
        suffixes = path.suffixes
        if self.path_has_number(path):
            suffixes.pop(self.NUMBER_SUFFIX_POSITION)
        suffixes.insert(self.NUMBER_SUFFIX_POSITION, f".{number}")
        stem = path.name.split(".", maxsplit=1)[0]
        new_name = stem + "".join(suffixes)
        return new_name

    def file_sort_key(self, path: Path) -> int:
        """Key function used to sort files"""
        return self.get_path_number(path) if self.path_has_number(path) else 0

    def get_logfiles(self) -> list[Path]:
        """Get the log files"""
        logfiles = list(filter(self.path_is_logfile, self.filename.parent.iterdir()))
        logfiles.sort(key=self.file_sort_key, reverse=True)
        return logfiles

    def rotate_file(self, path: Path) -> None:
        """Rotate a file's number suffix and remove it if it's too old"""

        # If uncompressed, compress
        if not path.name.endswith(".xz"):
            try:
                with open(path, "r", encoding="utf-8") as original_file:
                    original_data = original_file.read()
            except UnicodeDecodeError:
                # If the file is corrupted, throw it away
                path.unlink()
                return

            # Compress the file
            compressed_path = path.with_suffix(path.suffix + ".xz")
            with lzma.open(
                compressed_path,
                "wt",
                format=FORMAT_XZ,
                preset=PRESET_DEFAULT,
                encoding="utf-8",
            ) as lzma_file:
                lzma_file.write(original_data)
            path.unlink()
            path = compressed_path

        # Rename with new number suffix
        new_number = self.get_path_number(path) + 1
        new_path_name = self.set_path_number(path, new_number)
        path = path.rename(path.with_name(new_path_name))

        # Remove older files
        if new_number > self.backup_count:
            path.unlink()
            return

    def rotate(self) -> None:
        """Rotate the numbered suffix on the log files and remove old ones"""
        for path in self.get_logfiles():
            self.rotate_file(path)

    def __init__(self, filename: PathLike, backup_count: int = 2) -> None:
        self.filename = Path(filename)
        self.backup_count = backup_count
        self.create_dir()
        self.rotate()
        self.log_file = open(self.filename, "w", encoding="utf-8")
        shared.log_files = self.get_logfiles()
        super().__init__(self.log_file)

    def close(self) -> None:
        if self.log_file:
            self.log_file.close()
        super().close()



# steamgriddb.py
#
# Copyright 2022-2023 kramo
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from pathlib import Path
from typing import Any

import requests
from gi.repository import Gio
from requests.exceptions import HTTPError

from cartridges import shared
from cartridges.game import Game
from cartridges.utils.save_cover import convert_cover, save_cover


class SgdbError(Exception):
    pass


class SgdbAuthError(SgdbError):
    pass


class SgdbGameNotFound(SgdbError):
    pass


class SgdbBadRequest(SgdbError):
    pass


class SgdbNoImageFound(SgdbError):
    pass


class SgdbHelper:
    """Helper class to make queries to SteamGridDB"""

    base_url = "https://www.steamgriddb.com/api/v2/"

    @property
    def auth_headers(self) -> dict[str, str]:
        key = shared.schema.get_string("sgdb-key")
        headers = {"Authorization": f"Bearer {key}"}
        return headers

    def get_game_id(self, game: Game) -> Any:
        """Get grid results for a game. Can raise an exception."""
        uri = f"{self.base_url}search/autocomplete/{game.name}"
        res = requests.get(uri, headers=self.auth_headers, timeout=5)
        match res.status_code:
            case 200:
                return res.json()["data"][0]["id"]
            case 401:
                raise SgdbAuthError(res.json()["errors"][0])
            case 404:
                raise SgdbGameNotFound(res.status_code)
            case _:
                res.raise_for_status()

    def get_image_uri(self, game_id: str, animated: bool = False) -> Any:
        """Get the image for a SGDB game id"""
        uri = f"{self.base_url}grids/game/{game_id}?dimensions=600x900"
        if animated:
            uri += "&types=animated"
        res = requests.get(uri, headers=self.auth_headers, timeout=5)
        match res.status_code:
            case 200:
                data = res.json()["data"]
                if len(data) == 0:
                    raise SgdbNoImageFound()
                return data[0]["url"]
            case 401:
                raise SgdbAuthError(res.json()["errors"][0])
            case 404:
                raise SgdbGameNotFound(res.status_code)
            case _:
                res.raise_for_status()

    def conditionaly_update_cover(self, game: Game) -> None:
        """Update the game's cover if appropriate"""

        # Obvious skips
        use_sgdb = shared.schema.get_boolean("sgdb")
        if not use_sgdb or game.blacklisted:
            return

        image_trunk = shared.covers_dir / game.game_id
        still = image_trunk.with_suffix(".tiff")
        animated = image_trunk.with_suffix(".gif")
        prefer_sgdb = shared.schema.get_boolean("sgdb-prefer")

        # Do nothing if file present and not prefer SGDB
        if not prefer_sgdb and (still.is_file() or animated.is_file()):
            return

        # Get ID for the game
        try:
            sgdb_id = self.get_game_id(game)
        except (HTTPError, SgdbError) as error:
            logging.warning(
                "%s while getting SGDB ID for %s", type(error).__name__, game.name
            )
            raise error

        # Build different SGDB options to try
        image_uri_kwargs_sets = [{"animated": False}]
        if shared.schema.get_boolean("sgdb-animated"):
            image_uri_kwargs_sets.insert(0, {"animated": True})

        # Download covers
        for uri_kwargs in image_uri_kwargs_sets:
            try:
                uri = self.get_image_uri(sgdb_id, **uri_kwargs)
                response = requests.get(uri, timeout=5)
                tmp_file = Gio.File.new_tmp()[0]
                tmp_file_path = tmp_file.get_path()
                Path(tmp_file_path).write_bytes(response.content)
                tmp_cover_path = convert_cover(tmp_file_path)
                if tmp_cover_path:
                    save_cover(game.game_id, tmp_cover_path)
                    tmp_cover_path.unlink(missing_ok=True)
                tmp_file_path.unlink(missing_ok=True)
            except SgdbAuthError as error:
                # Let caller handle auth errors
                raise error
            except (HTTPError, SgdbError) as error:
                logging.warning(
                    "%s while getting image for %s kwargs=%s",
                    type(error).__name__,
                    game.name,
                    str(uri_kwargs),
                )
                continue
            else:
                # Stop as soon as one is finished
                return

        # No image was added
        logging.warning(
            'No matching image found for game "%s" (SGDB ID %d)',
            game.name,
            sgdb_id,
        )
        raise SgdbNoImageFound()



# relative_date.py
#
# Copyright 2022-2023 kramo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
from typing import Any

from gi.repository import GLib


def relative_date(timestamp: int) -> Any:  # pylint: disable=too-many-return-statements
    days_no = ((today := datetime.today()) - datetime.fromtimestamp(timestamp)).days

    if days_no == 0:
        return _("Today")
    if days_no == 1:
        return _("Yesterday")
    if days_no <= (day_of_week := today.weekday()):
        return GLib.DateTime.new_from_unix_utc(timestamp).format("%A")
    if days_no <= day_of_week + 7:
        return _("Last Week")
    if days_no <= (day_of_month := today.day):
        return _("This Month")
    if days_no <= day_of_month + 30:
        return _("Last Month")
    if days_no < (day_of_year := today.timetuple().tm_yday):
        return GLib.DateTime.new_from_unix_utc(timestamp).format("%B")
    if days_no <= day_of_year + 365:
        return _("Last Year")
    return GLib.DateTime.new_from_unix_utc(timestamp).format("%Y")



# itch_source.py
#
# Copyright 2022-2023 kramo
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from shutil import rmtree
from sqlite3 import connect
from typing import NamedTuple

from cartridges import shared
from cartridges.game import Game
from cartridges.importer.location import Location, LocationSubPath
from cartridges.importer.source import SourceIterable, URLExecutableSource
from cartridges.utils.sqlite import copy_db


class ItchSourceIterable(SourceIterable):
    source: "ItchSource"

    def __iter__(self):
        """Generator method producing games"""

        # Query the database
        db_request = """
        SELECT
            games.id,
            games.title,
            games.cover_url,
            games.still_cover_url,
            caves.id
        FROM
            'caves'
        INNER JOIN
            'games'
        ON
            caves.game_id = games.id
        ;
        """
        db_path = copy_db(self.source.locations.config["butler.db"])
        connection = connect(db_path)
        cursor = connection.execute(db_request)

        # Create games from the db results
        for row in cursor:
            values = {
                "added": shared.import_time,
                "source": self.source.source_id,
                "name": row[1],
                "game_id": self.source.game_id_format.format(game_id=row[0]),
                "executable": self.source.make_executable(cave_id=row[4]),
            }
            additional_data = {"online_cover_url": row[3] or row[2]}
            game = Game(values)
            yield (game, additional_data)

        # Cleanup
        rmtree(str(db_path.parent))


class ItchLocations(NamedTuple):
    config: Location


class ItchSource(URLExecutableSource):
    source_id = "itch"
    name = _("itch")
    iterable_class = ItchSourceIterable
    url_format = "itch://caves/{cave_id}/launch"
    available_on = {"linux", "win32", "darwin"}

    locations: ItchLocations

    def __init__(self) -> None:
        super().__init__()
        self.locations = ItchLocations(
            Location(
                schema_key="itch-location",
                candidates=(
                    shared.flatpak_dir / "io.itch.itch" / "config" / "itch",
                    shared.config_dir / "itch",
                    shared.host_config_dir / "itch",
                    shared.appdata_dir / "itch",
                    shared.app_support_dir / "itch",
                ),
                paths={
                    "butler.db": LocationSubPath("db/butler.db"),
                },
                invalid_subtitle=Location.CONFIG_INVALID_SUBTITLE,
            )
        )



# flatpak_source.py
#
# Copyright 2022-2023 kramo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from itertools import chain
from pathlib import Path
from typing import NamedTuple

from gi.repository import GLib, Gtk

from cartridges import shared
from cartridges.game import Game
from cartridges.importer.location import Location, LocationSubPath
from cartridges.importer.source import ExecutableFormatSource, SourceIterable


class FlatpakSourceIterable(SourceIterable):
    source: "FlatpakSource"

    def __iter__(self):
        """Generator method producing games"""

        icon_theme = Gtk.IconTheme.new()
        if user_data := self.source.locations.user_data["icons"]:
            icon_theme.add_search_path(str(user_data))

        if system_data := self.source.locations.system_data["icons"]:
            icon_theme.add_search_path(str(system_data))

        if not (system_data or user_data):
            return

        blacklist = (
            {
                "hu.kramo.Cartridges",
                "hu.kramo.Cartridges.Devel",
                "page.kramo.Cartridges",
                "page.kramo.Cartridges.Devel",
            }
            if shared.schema.get_boolean("flatpak-import-launchers")
            else {
                "hu.kramo.Cartridges",
                "hu.kramo.Cartridges.Devel",
                "page.kramo.Cartridges",
                "page.kramo.Cartridges.Devel",
                "com.valvesoftware.Steam",
                "net.lutris.Lutris",
                "com.heroicgameslauncher.hgl",
                "com.usebottles.Bottles",
                "io.itch.itch",
                "org.libretro.RetroArch",
            }
        )

        generators = set(
            location.iterdir()
            for location in (
                self.source.locations.user_data["applications"],
                self.source.locations.system_data["applications"],
            )
            if location
        )

        for entry in chain(*generators):
            if entry.suffix != ".desktop":
                continue

            keyfile = GLib.KeyFile.new()

            try:
                keyfile.load_from_file(str(entry), 0)

                if "Game" not in keyfile.get_string_list("Desktop Entry", "Categories"):
                    continue

                if (
                    flatpak_id := keyfile.get_string("Desktop Entry", "X-Flatpak")
                ) in blacklist or flatpak_id != entry.stem:
                    continue

                name = keyfile.get_string("Desktop Entry", "Name")

            except GLib.Error:
                continue

            values = {
                "source": self.source.source_id,
                "added": shared.import_time,
                "name": name,
                "game_id": self.source.game_id_format.format(game_id=flatpak_id),
                "executable": self.source.make_executable(flatpak_id=flatpak_id),
            }
            game = Game(values)

            additional_data = {}

            try:
                if (
                    icon_path := icon_theme.lookup_icon(
                        keyfile.get_string("Desktop Entry", "Icon"),
                        None,
                        512,
                        1,
                        shared.win.get_direction(),
                        0,
                    )
                    .get_file()
                    .get_path()
                ):
                    additional_data = {"local_icon_path": Path(icon_path)}
                else:
                    pass
            except GLib.Error:
                pass

            yield (game, additional_data)


class FlatpakLocations(NamedTuple):
    system_data: Location
    user_data: Location


class FlatpakSource(ExecutableFormatSource):
    """Generic Flatpak source"""

    source_id = "flatpak"
    name = _("Flatpak")
    iterable_class = FlatpakSourceIterable
    executable_format = "flatpak run {flatpak_id}"
    available_on = {"linux"}

    locations: FlatpakLocations

    def __init__(self) -> None:
        super().__init__()
        self.locations = FlatpakLocations(
            Location(
                schema_key="flatpak-system-location",
                candidates=("/var/lib/flatpak/",),
                paths={
                    "applications": LocationSubPath("exports/share/applications", True),
                    "icons": LocationSubPath("exports/share/icons", True),
                },
                invalid_subtitle=Location.DATA_INVALID_SUBTITLE,
                optional=True,
            ),
            Location(
                schema_key="flatpak-user-location",
                candidates=(shared.data_dir / "flatpak",),
                paths={
                    "applications": LocationSubPath("exports/share/applications", True),
                    "icons": LocationSubPath("exports/share/icons", True),
                },
                invalid_subtitle=Location.DATA_INVALID_SUBTITLE,
                optional=True,
            ),
        )



# steam_source.py
#
# Copyright 2022-2023 kramo
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import re
from pathlib import Path
from typing import Iterable, NamedTuple

from cartridges import shared
from cartridges.game import Game
from cartridges.importer.location import Location, LocationSubPath
from cartridges.importer.source import SourceIterable, URLExecutableSource
from cartridges.utils.steam import SteamFileHelper, SteamInvalidManifestError


class SteamSourceIterable(SourceIterable):
    source: "SteamSource"

    def get_manifest_dirs(self) -> Iterable[Path]:
        """Get dirs that contain Steam app manifests"""
        libraryfolders_path = self.source.locations.data["libraryfolders.vdf"]
        with open(libraryfolders_path, "r", encoding="utf-8") as file:
            contents = file.read()
        return [
            Path(path) / "steamapps"
            for path in re.findall('"path"\\s+"(.*)"\n', contents, re.IGNORECASE)
        ]

    def get_manifests(self) -> Iterable[Path]:
        """Get app manifests"""
        manifests = set()
        for steamapps_dir in self.get_manifest_dirs():
            if not steamapps_dir.is_dir():
                continue
            manifests.update(
                [
                    manifest
                    for manifest in steamapps_dir.glob("appmanifest_*.acf")
                    if manifest.is_file()
                ]
            )
        return manifests

    def __iter__(self):
        """Generator method producing games"""
        appid_cache = set()
        manifests = self.get_manifests()

        for manifest in manifests:
            # Get metadata from manifest
            steam = SteamFileHelper()
            try:
                local_data = steam.get_manifest_data(manifest)
            except (OSError, SteamInvalidManifestError) as error:
                logging.debug("Couldn't load appmanifest %s", manifest, exc_info=error)
                continue

            # Skip non installed games
            installed_mask = 4
            if not int(local_data["stateflags"]) & installed_mask:
                logging.debug("Skipped %s: not installed", manifest)
                continue

            # Skip duplicate appids
            appid = local_data["appid"]
            if appid in appid_cache:
                logging.debug("Skipped %s: appid already seen during import", manifest)
                continue
            appid_cache.add(appid)

            # Build game from local data
            values = {
                "added": shared.import_time,
                "name": local_data["name"],
                "source": self.source.source_id,
                "game_id": self.source.game_id_format.format(game_id=appid),
                "executable": self.source.make_executable(game_id=appid),
            }
            game = Game(values)

            # Add official cover image
            image_path = (
                self.source.locations.data["librarycache"]
                / f"{appid}_library_600x900.jpg"
            )
            additional_data = {"local_image_path": image_path, "steam_appid": appid}

            yield (game, additional_data)


class SteamLocations(NamedTuple):
    data: Location


class SteamSource(URLExecutableSource):
    source_id = "steam"
    name = _("Steam")
    available_on = {"linux", "win32", "darwin"}
    iterable_class = SteamSourceIterable
    url_format = "steam://rungameid/{game_id}"

    locations: SteamLocations

    def __init__(self) -> None:
        super().__init__()
        self.locations = SteamLocations(
            Location(
                schema_key="steam-location",
                candidates=(
                    shared.home / ".steam" / "steam",
                    shared.data_dir / "Steam",
                    shared.flatpak_dir / "com.valvesoftware.Steam" / "data" / "Steam",
                    shared.programfiles32_dir / "Steam",
                    shared.app_support_dir / "Steam",
                ),
                paths={
                    "libraryfolders.vdf": LocationSubPath(
                        "steamapps/libraryfolders.vdf"
                    ),
                    "librarycache": LocationSubPath("appcache/librarycache", True),
                },
                invalid_subtitle=Location.DATA_INVALID_SUBTITLE,
            )
        )



# desktop_source.py
#
# Copyright 2023 kramo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import shlex
import subprocess
from pathlib import Path
from typing import NamedTuple

from gi.repository import GLib, Gtk

from cartridges import shared
from cartridges.game import Game
from cartridges.importer.source import Source, SourceIterable


class DesktopSourceIterable(SourceIterable):
    source: "DesktopSource"

    def __iter__(self):
        """Generator method producing games"""

        icon_theme = Gtk.IconTheme.new()

        search_paths = [
            shared.host_data_dir,
            "/run/host/usr/local/share",
            "/run/host/usr/share",
            "/run/host/usr/share/pixmaps",
            "/usr/share/pixmaps",
        ] + GLib.get_system_data_dirs()

        for search_path in search_paths:
            path = Path(search_path)

            if not str(search_path).endswith("/pixmaps"):
                path = path / "icons"

            if not path.is_dir():
                continue

            if str(path).startswith("/app/"):
                continue

            icon_theme.add_search_path(str(path))

        launch_command, full_path = self.check_launch_commands()

        for path in search_paths:
            if str(path).startswith("/app/"):
                continue

            path = Path(path) / "applications"

            if not path.is_dir():
                continue

            for entry in path.iterdir():
                if entry.suffix != ".desktop":
                    continue

                # Skip Lutris games
                if str(entry.name).startswith("net.lutris."):
                    continue

                keyfile = GLib.KeyFile.new()

                try:
                    keyfile.load_from_file(str(entry), 0)

                    if "Game" not in keyfile.get_string_list(
                        "Desktop Entry", "Categories"
                    ):
                        continue

                    name = keyfile.get_string("Desktop Entry", "Name")
                    executable = keyfile.get_string("Desktop Entry", "Exec").split(
                        " %"
                    )[0]
                except GLib.Error:
                    continue

                try:
                    try_exec = "which " + keyfile.get_string("Desktop Entry", "TryExec")
                    if not self.check_command(try_exec):
                        continue

                except GLib.Error:
                    pass

                # Skip Steam games
                if "steam://rungameid/" in executable:
                    continue

                # Skip Heroic games
                if "heroic://launch/" in executable:
                    continue

                # Skip Bottles games
                if "bottles-cli " in executable:
                    continue

                try:
                    if keyfile.get_boolean("Desktop Entry", "NoDisplay"):
                        continue
                except GLib.Error:
                    pass

                try:
                    if keyfile.get_boolean("Desktop Entry", "Hidden"):
                        continue
                except GLib.Error:
                    pass

                # Strip /run/host from Flatpak paths
                if entry.is_relative_to(prefix := "/run/host"):
                    entry = Path("/") / entry.relative_to(prefix)

                launch_arg = shlex.quote(str(entry if full_path else entry.stem))

                values = {
                    "source": self.source.source_id,
                    "added": shared.import_time,
                    "name": name,
                    "game_id": f"desktop_{entry.stem}",
                    "executable": f"{launch_command} {launch_arg}",
                }
                game = Game(values)

                additional_data = {}

                try:
                    icon_str = keyfile.get_string("Desktop Entry", "Icon")
                except GLib.Error:
                    yield game
                    continue
                else:
                    if "/" in icon_str:
                        additional_data = {"local_icon_path": Path(icon_str)}
                        yield (game, additional_data)
                        continue

                try:
                    if (
                        icon_path := icon_theme.lookup_icon(
                            icon_str,
                            None,
                            512,
                            1,
                            shared.win.get_direction(),
                            0,
                        )
                        .get_file()
                        .get_path()
                    ):
                        additional_data = {"local_icon_path": Path(icon_path)}
                except GLib.Error:
                    pass

                yield (game, additional_data)

    def check_command(self, command) -> bool:
        flatpak_str = "flatpak-spawn --host /bin/sh -c "

        if os.getenv("FLATPAK_ID") == shared.APP_ID:
            command = flatpak_str + shlex.quote(command)

        try:
            subprocess.run(command, shell=True, check=True)
        except subprocess.CalledProcessError:
            return False

        return True

    def check_launch_commands(self) -> (str, bool):
        """Check whether `gio launch` `gtk4-launch` or `gtk-launch` are available on the system"""
        commands = (("gio launch", True), ("gtk4-launch", False), ("gtk-launch", False))

        for command, full_path in commands:
            # Even if `gio` is available, `gio launch` is only available on GLib >= 2.67.2
            command_to_check = (
                "gio help launch" if command == "gio launch" else f"which {command}"
            )

            if self.check_command(command_to_check):
                return command, full_path

        return commands[2]


class DesktopLocations(NamedTuple):
    pass


class DesktopSource(Source):
    """Generic Flatpak source"""

    source_id = "desktop"
    name = _("Desktop Entries")
    iterable_class = DesktopSourceIterable
    available_on = {"linux"}

    locations: DesktopLocations

    def __init__(self) -> None:
        super().__init__()
        self.locations = DesktopLocations()



# steam_api_manager.py
#
# Copyright 2023 Geoffrey Coulaud
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from requests.exceptions import HTTPError, SSLError
from urllib3.exceptions import ConnectionError as Urllib3ConnectionError

from cartridges.game import Game
from cartridges.store.managers.async_manager import AsyncManager
from cartridges.utils.steam import (
    SteamAPIHelper,
    SteamGameNotFoundError,
    SteamNotAGameError,
    SteamRateLimiter,
)


class SteamAPIManager(AsyncManager):
    """Manager in charge of completing a game's data from the Steam API"""

    retryable_on = (HTTPError, SSLError, Urllib3ConnectionError)

    steam_api_helper: SteamAPIHelper = None
    steam_rate_limiter: SteamRateLimiter = None

    def __init__(self) -> None:
        super().__init__()
        self.steam_rate_limiter = SteamRateLimiter()
        self.steam_api_helper = SteamAPIHelper(self.steam_rate_limiter)

    def main(self, game: Game, additional_data: dict) -> None:
        # Skip non-Steam games
        appid = additional_data.get("steam_appid", None)
        if appid is None:
            return
        # Get online metadata
        try:
            online_data = self.steam_api_helper.get_api_data(appid=appid)
        except (SteamNotAGameError, SteamGameNotFoundError):
            game.update_values({"blacklisted": True})
        else:
            game.update_values(online_data)
